@@ -12,11 +12,19 @@ import re
 import datetime
 from lxml import etree
 
+try:
+    from urllib.parse import unquote
+except ImportError:
+    from urllib import unquote
+
 from caldav.lib import error, vcal, url
 from caldav.lib.url import URL
 from caldav.elements import dav, cdav
 from caldav.lib.python_utilities import to_unicode
 
+## utility for formatting a response xml tree to an error string
+def errmsg(r):
+    return "%s %s\n\n%s" % (r.status, r.reason, r.raw)
 
 class DAVObject(object):
     """
@@ -68,19 +76,21 @@ class DAVObject(object):
         depth = 1
         properties = {}
 
-        props = [dav.ResourceType(), ]
+        props = [dav.ResourceType(), dav.DisplayName()]
         response = self._query_properties(props, depth)
         properties = self._handle_prop_response(response=response, props=props, type=type, what='tag')
 
         for path in list(properties.keys()):
             resource_type = properties[path][dav.ResourceType.tag]
+            resource_name = properties[path][dav.DisplayName.tag]
+
             if resource_type == type or type is None:
-                ## TODO: investigate the RFCs thoroughly - why does a "get 
-                ## members of this collection"-request also return the collection URL itself?  
-                ## And why is the strip_trailing_slash-method needed?  The collection URL 
+                ## TODO: investigate the RFCs thoroughly - why does a "get
+                ## members of this collection"-request also return the collection URL itself?
+                ## And why is the strip_trailing_slash-method needed?  The collection URL
                 ## should always end with a slash according to RFC 2518, section 5.2.
                 if self.url.strip_trailing_slash() != self.url.join(path).strip_trailing_slash():
-                    c.append((self.url.join(path), resource_type))
+                    c.append((self.url.join(path), resource_type, resource_name))
 
         return c
 
@@ -106,6 +116,9 @@ class DAVObject(object):
         """
         if url is None:
             url = self.url
+            if not url.endswith('/'):
+                url = URL(str(url) + '/')
+
         body = ""
         if root:
             body = etree.tostring(root.xmlelement(), encoding="utf-8",
@@ -113,13 +126,13 @@ class DAVObject(object):
         ret = getattr(self.client, query_method)(
             url, body, depth)
         if ret.status == 404:
-            raise error.NotFoundError(ret.raw)
+            raise error.NotFoundError(errmsg(ret))
         if (
                 (expected_return_value is not None and ret.status != expected_return_value) or
                 ret.status >= 400):
-            raise error.exception_by_method[query_method](ret.raw)
+            raise error.exception_by_method[query_method](errmsg(ret))
         return ret
-        
+
 
     def _handle_prop_response(self, response, props=[], type=None, what='text'):
         """
@@ -131,18 +144,15 @@ class DAVObject(object):
         # All items should be in a <D:response> element
         for r in response.tree.findall('.//' + dav.Response.tag):
             status = r.find('.//' + dav.Status.tag)
-            if not status.text.endswith('200 OK'):
-                if status.text.endswith('404 Not Found'):
-                    ## Silently ignoring the 404
-                    ## (we should probably spit out some debug logging here)
-                    continue
-                raise error.ReportError(response.raw) ## TODO: may be wrong error class
-
-            href = r.find('.//' + dav.Href.tag).text
+            if not ' 200 ' in status.text and not ' 207 ' in status.text and not ' 404 ' in status.text:
+                raise error.ReportError(errmsg(response)) ## TODO: may be wrong error class
+            href = unquote(r.find('.//' + dav.Href.tag).text)
             properties[href] = {}
             for p in props:
                 t = r.find(".//" + p.tag)
-                if len(list(t)) > 0:
+                if t is None:
+                    val = None
+                elif t is not None and list(t):
                     if type is not None:
                         val = t.find(".//" + type)
                     else:
@@ -188,7 +198,6 @@ class DAVObject(object):
         """
         Set properties (PROPPATCH) for this object.
 
-        Parameters:
          * props = [dav.DisplayName('name'), ...]
 
         Returns:
@@ -202,8 +211,8 @@ class DAVObject(object):
 
         statuses = r.tree.findall(".//" + dav.Status.tag)
         for s in statuses:
-            if not s.text.endswith("200 OK"):
-                raise error.PropsetError(r.raw)
+            if not ' 200 ' in s.text:
+                raise error.PropsetError(errmsg(r))
 
         return self
 
@@ -226,7 +235,7 @@ class DAVObject(object):
 
             #TODO: find out why we get 404
             if r.status not in (200, 204, 404):
-                raise error.DeleteError(r.raw)
+                raise error.DeleteError(errmsg(r))
 
     def __str__(self):
         return str(self.url)
@@ -246,15 +255,15 @@ class CalendarSet(DAVObject):
         cals = []
 
         data = self.children(cdav.Calendar.tag)
-        for c_url, c_type in data:
-            cals.append(Calendar(self.client, c_url, parent=self))
+        for c_url, c_type, c_name in data:
+            cals.append(Calendar(self.client, c_url, parent=self, name=c_name))
 
         return cals
 
     def make_calendar(self, name=None, cal_id=None, supported_calendar_component_set=None):
         """
         Utility method for creating a new calendar.
-        
+
         Parameters:
          * name: the name of the new calendar
          * cal_id: the uuid of the new calendar
@@ -299,7 +308,7 @@ class Principal(DAVObject):
         self.client = client
         self._calendar_home_set = None
 
-        ## backwards compatibility.  
+        ## backwards compatibility.
         if url is not None:
             self.url = client.url.join(URL.objectify(url))
         else:
@@ -360,15 +369,16 @@ class Calendar(DAVObject):
         path = self.parent.url.join(id)
         self.url = path
 
-        ## TODO: mkcalendar seems to ignore the body on most servers?  
+        ## TODO: mkcalendar seems to ignore the body on most servers?
         ## at least the name doesn't get set this way.
         ## zimbra gives 500 (!) if body is omitted ...
 
-        cal = cdav.CalendarCollection()
-        coll = dav.Collection() + cal
-        type = dav.ResourceType() + coll
+        ## ehm ... this element seems non-existent in the RFC?  Breaks with baikal, too ...
+        #cal = cdav.CalendarCollection()
+        #coll = dav.Collection() # + cal ## also breaks on baikal, and probably not needed?
+        #type = dav.ResourceType() ## probably not needed?
 
-        prop = dav.Prop() + [type,]
+        prop = dav.Prop() #+ [type,]
         if name:
             display_name = dav.DisplayName(name)
             prop += [display_name,]
@@ -396,7 +406,7 @@ class Calendar(DAVObject):
         ## event uuid matches the event url, and return either 201 or
         ## 302 - but alas, try to do a GET towards the event and we
         ## get 404!  But turn around and replace the calendar ID with
-        ## the calendar name in the URL and hey ... it works!  
+        ## the calendar name in the URL and hey ... it works!
 
         ## TODO: write test cases for calendars with non-trivial
         ## names and calendars with names already matching existing
@@ -406,12 +416,15 @@ class Calendar(DAVObject):
             ret = self.client.request(zimbra_url)
             if ret.status == 404:
                 raise error.NotFoundError
-            ## insane server
-            self.url = zimbra_url
+            ## special hack for radicale.  It will happily accept any calendar-URL without returning 404.
+            ret = self.client.request(self.parent.url.join('ANYTHINGGOESHEREthisshouldforsurereturn404'))
+            if ret.status == 404:
+                ## insane server
+                self.url = zimbra_url
         except error.NotFoundError:
             ## sane server
             pass
-        
+
     def add_event(self, ical):
         """
         Add a new event to the calendar, with the given ical.
@@ -453,7 +466,7 @@ class Calendar(DAVObject):
                 self.url = URL.objectify(str(self.url) + '/')
         return self
 
-    def date_search(self, start, end=None):
+    def date_search(self, start, end=None, compfilter="VEVENT"):
         """
         Search events by date in the calendar. Recurring events are
         expanded if they are occuring during the specified time frame
@@ -462,15 +475,16 @@ class Calendar(DAVObject):
         Parameters:
          * start = datetime.today().
          * end = same as above.
+         * compfilter = defaults to events only.  Set to None to fetch all calendar components.
 
         Returns:
-         * [Event(), ...]
+         * [CalendarObjectResource(), ...]
 
         """
         matches = []
 
         # build the request
-        
+
         ## Some servers will raise an error if we send the expand flag
         ## but don't set any end-date - expand doesn't make much sense
         ## if we have one recurring event describing an indefinite
@@ -482,9 +496,10 @@ class Calendar(DAVObject):
             data = cdav.CalendarData()
         prop = dav.Prop() + data
 
-        range = cdav.TimeRange(start, end)
-        vevent = cdav.CompFilter("VEVENT") + range
-        vcalendar = cdav.CompFilter("VCALENDAR") + vevent
+        query = cdav.TimeRange(start, end)
+        if compfilter:
+            query = cdav.CompFilter(compfilter) + query
+        vcalendar = cdav.CompFilter("VCALENDAR") + query
         filter = cdav.Filter() + vcalendar
 
         root = cdav.CalendarQuery() + [prop, filter]
@@ -493,7 +508,7 @@ class Calendar(DAVObject):
         for r in results:
             matches.append(
                 Event(self.client, url=self.url.join(r), data=results[r][cdav.CalendarData.tag], parent=self))
-        
+
         return matches
 
     def freebusy_request(self, start, end):
@@ -517,13 +532,14 @@ class Calendar(DAVObject):
         fetches a list of journal entries.
         """
 
-    def todos(self, sort_key='due', include_completed=False):
+    def todos(self, sort_keys=('due','priority'), include_completed=False, sort_key=None):
         """
         fetches a list of todo events.
 
         Parameters:
-         * sort_key: use this field in the VTODO for sorting (lower case string, i.e. 'priority').
+         * sort_keys: use this field in the VTODO for sorting (iterable of lower case string, i.e. ('priority','due')).
          * include_completed: boolean - by default, only pending tasks are listed
+         * sort_key: DEPRECATED, for backwards compatibility with version 0.4.
         """
         ## ref https://www.ietf.org/rfc/rfc4791.txt, section 7.8.9
         matches = []
@@ -532,12 +548,16 @@ class Calendar(DAVObject):
         data = cdav.CalendarData()
         prop = dav.Prop() + data
 
+        if sort_key:
+            sort_keys = (sort_key,)
+
         if not include_completed:
             vnotcompleted = cdav.TextMatch('COMPLETED', negate=True)
             vnotcancelled = cdav.TextMatch('CANCELLED', negate=True)
-            vstatus = cdav.PropFilter('STATUS') + vnotcancelled + vnotcompleted
+            vstatusNotCompleted = cdav.PropFilter('STATUS') + vnotcompleted
+            vstatusNotCancelled = cdav.PropFilter('STATUS') + vnotcancelled
             vnocompletedate = cdav.PropFilter('COMPLETED') + cdav.NotDefined()
-            vtodo = cdav.CompFilter("VTODO") + vnocompletedate + vstatus
+            vtodo = cdav.CompFilter("VTODO") + vnocompletedate + vstatusNotCompleted + vstatusNotCancelled
         else:
             vtodo = cdav.CompFilter("VTODO")
         vcalendar = cdav.CompFilter("VCALENDAR") + vtodo
@@ -552,24 +572,47 @@ class Calendar(DAVObject):
                 Todo(self.client, url=self.url.join(r), data=results[r][cdav.CalendarData.tag], parent=self))
 
         def sort_key_func(x):
-            val = getattr(x.instance.vtodo, sort_key, None)
-            if not val:
-                return None
-            val = val.value
-            if hasattr(val, 'strftime'):
-                return val.strftime('%F%H%M%S')
-            return val
-        if sort_key:
+            ret = []
+            defaults = {
+                'due': '2050-01-01',
+                'dtstart': '1970-01-01',
+                'priority': '0',
+                'isnt_overdue': not (hasattr(x.instance.vtodo, 'due') and x.instance.vtodo.due.value.strftime('%F%H%M%S') < datetime.datetime.now().strftime('%F%H%M%S')),
+                'hasnt_started': (hasattr(x.instance.vtodo, 'dtstart') and x.instance.vtodo.dtstart.value.strftime('%F%H%M%S') > datetime.datetime.now().strftime('%F%H%M%S'))
+            }
+            for sort_key in sort_keys:
+                val = getattr(x.instance.vtodo, sort_key, None)
+                if val is None:
+                    ret.append(defaults.get(sort_key,'0'))
+                    continue
+                val = val.value
+                if hasattr(val, 'strftime'):
+                    ret.append(val.strftime('%F%H%M%S'))
+                else:
+                    ret.append(val)
+            return ret
+        if sort_keys:
             matches.sort(key=sort_key_func)
         return matches
 
+    def _calendar_comp_class_by_data(self, data):
+        for line in data.split('\n'):
+            if line == 'BEGIN:VEVENT':
+                return Event
+            if line == 'BEGIN:VTODO':
+                return Todo
+            if line == 'BEGIN:VJOURNAL':
+                return Journal
+            if line == 'BEGIN:VFREEBUSY':
+                return FreeBusy
+
     def event_by_url(self, href, data=None):
         """
         Returns the event with the given URL
         """
         return Event(url=href, data=data, parent=self).load()
 
-    def event_by_uid(self, uid):
+    def object_by_uid(self, uid, comp_filter=None):
         """
         Get one event from the calendar.
 
@@ -582,10 +625,11 @@ class Calendar(DAVObject):
         data = cdav.CalendarData()
         prop = dav.Prop() + data
 
-        match = cdav.TextMatch(uid)
-        propf = cdav.PropFilter("UID") + match
-        vevent = cdav.CompFilter("VEVENT") + propf
-        vcalendar = cdav.CompFilter("VCALENDAR") + vevent
+        query = cdav.TextMatch(uid)
+        query = cdav.PropFilter("UID") + query
+        if comp_filter:
+            query = comp_filter + query
+        vcalendar = cdav.CompFilter("VCALENDAR") + query
         filter = cdav.Filter() + vcalendar
 
         root = cdav.CalendarQuery() + [prop, filter]
@@ -593,18 +637,23 @@ class Calendar(DAVObject):
         response = self._query(root, 1, 'report')
 
         if response.status == 404:
-            raise error.NotFoundError(response.raw)
+            raise error.NotFoundError(errmsg(response))
         elif response.status == 400:
-            raise error.ReportError(response.raw)
-            
+            raise error.ReportError(errmsg(response))
+
         r = response.tree.find(".//" + dav.Response.tag)
         if r is not None:
-            href = r.find(".//" + dav.Href.tag).text
-            data = r.find(".//" + cdav.CalendarData.tag).text
-            return Event(self.client, url=URL.objectify(href), data=data, parent=self)
+            href = unquote(r.find(".//" + dav.Href.tag).text)
+            data = unquote(r.find(".//" + cdav.CalendarData.tag).text)
+            return self._calendar_comp_class_by_data(data)(self.client, url=URL.objectify(href), data=data, parent=self)
         else:
-            raise error.NotFoundError(response.raw)
+            raise error.NotFoundError(errmsg(response))
 
+    def todo_by_uid(self, uid):
+        return self.object_by_uid(uid, comp_filter=cdav.CompFilter("VTODO"))
+
+    def event_by_uid(self, uid):
+        return self.object_by_uid(uid, comp_filter=cdav.CompFilter("VEVENT"))
     ## alias for backward compatibility
     event = event_by_uid
 
@@ -623,7 +672,7 @@ class Calendar(DAVObject):
         vcalendar = cdav.CompFilter("VCALENDAR") + vevent
         filter = cdav.Filter() + vcalendar
         root = cdav.CalendarQuery() + [prop, filter]
-        
+
         response = self._query(root, 1, query_method='report')
         results = self._handle_prop_response(response, props=[cdav.CalendarData()])
         for r in results:
@@ -649,7 +698,7 @@ class Calendar(DAVObject):
         vcalendar = cdav.CompFilter("VCALENDAR") + vevent
         filter = cdav.Filter() + vcalendar
         root = cdav.CalendarQuery() + [prop, filter]
-        
+
         response = self._query(root, 1, query_method='report')
         results = self._handle_prop_response(response, props=[cdav.CalendarData()])
         for r in results:
@@ -674,13 +723,22 @@ class CalendarObjectResource(DAVObject):
         if data is not None:
             self.data = data
 
+    def copy(self, keep_uid=False, new_parent=None):
+        """
+        Events, todos etc can be copied within the same calendar, to another calendar or even to another caldav server
+        """
+        return self.__class__(
+            parent=new_parent or self.parent,
+            data=self.data,
+            id = self.id if keep_uid else str(uuid.uuid1()))
+
     def load(self):
         """
         Load the object from the caldav server.
         """
         r = self.client.request(self.url)
         if r.status == 404:
-            raise error.NotFoundError(r.raw)
+            raise error.NotFoundError(errmsg(r))
         self.data = vcal.fix(r.raw)
         return self
 
@@ -688,9 +746,26 @@ class CalendarObjectResource(DAVObject):
         if id is None and path is not None and str(path).endswith('.ics'):
             id = re.search('(/|^)([^/]*).ics',str(path)).group(2)
         elif id is None:
-            for obj in ('vevent', 'vtodo', 'vjournal', 'vfreebusy'):
-                if hasattr(self.instance, obj):
-                    id = getattr(self.instance, obj).uid.value
+            for obj_type in ('vevent', 'vtodo', 'vjournal', 'vfreebusy'):
+                obj = None
+                if hasattr(self.instance, obj_type):
+                    obj =  getattr(self.instance, obj_type)
+                elif self.instance.name.lower() == obj_type:
+                    obj = self.instance
+                if obj is not None:
+                    id = obj.uid.value
+                    break
+        else:
+            for obj_type in ('vevent', 'vtodo', 'vjournal', 'vfreebusy'):
+                obj = None
+                if hasattr(self.instance, obj_type):
+                    obj = getattr(self.instance, obj_type)
+                elif self.instance.name.lower() == obj_type:
+                    obj = self.instance
+                if obj is not None:
+                    if not hasattr(obj, 'uid'):
+                        obj.add('uid')
+                    obj.uid = id
                     break
         if path is None:
             path = id + ".ics"
@@ -701,7 +776,7 @@ class CalendarObjectResource(DAVObject):
         if r.status == 302:
             path = [x[1] for x in r.headers if x[0]=='location'][0]
         elif not (r.status in (204, 201)):
-            raise error.PutError(r.raw)
+            raise error.PutError(errmsg(r))
 
         self.url = URL.objectify(path)
         self.id = id
@@ -723,7 +798,7 @@ class CalendarObjectResource(DAVObject):
 
     def _set_data(self, data):
         self._data = vcal.fix(data)
-        self._instance = vobject.readOne(io.StringIO(to_unicode(self._data)))
+        self._instance = vobject.readOne(to_unicode(self._data))
         return self
 
     def _get_data(self):
@@ -780,6 +855,10 @@ class Todo(CalendarObjectResource):
         """
         if not completion_timestamp:
             completion_timestamp = datetime.datetime.now()
+        if not hasattr(self.instance.vtodo, 'status'):
+            self.instance.vtodo.add('status')
         self.instance.vtodo.status.value = 'COMPLETED'
         self.instance.vtodo.add('completed').value = completion_timestamp
         self.save()
+
+
