@@ -13,6 +13,7 @@ import logging
 import threading
 import time
 from nose.tools import assert_equal, assert_not_equal, assert_raises
+from nose.plugins.skip import SkipTest
 
 from .conf import caldav_servers, proxy, proxy_noport
 from .proxy import ProxyHandler, NonThreadingHTTPServer
@@ -125,38 +126,35 @@ STATUS:NEEDS-ACTION
 END:VTODO
 END:VCALENDAR"""
 
+todo4 = """
+BEGIN:VCALENDAR
+VERSION:2.0
+PRODID:-//Example Corp.//CalDAV Client//EN
+BEGIN:VTODO
+UID:19970901T130000Z-123406@host.com
+DTSTAMP:19970901T130000Z
+SUMMARY:1996 Income Tax Preparation
+CLASS:CONFIDENTIAL
+CATEGORIES:FAMILY,FINANCE
+PRIORITY:1
+STATUS:NEEDS-ACTION
+END:VTODO
+END:VCALENDAR"""
+
 ## example from http://www.kanzaki.com/docs/ical/vjournal.html
-journal1 = """
+journal = """
 BEGIN:VCALENDAR
 VERSION:2.0
 PRODID:-//Example Corp.//CalDAV Client//EN
 BEGIN:VJOURNAL
-UID:19970901T130000Z-123405@host.com
-DTSTAMP:19970901T1300Z
+UID:19970901T130000Z-123405@example.com
+DTSTAMP:19970901T130000Z
 DTSTART;VALUE=DATE:19970317
 SUMMARY:Staff meeting minutes
 DESCRIPTION:1. Staff meeting: Participants include Joe\, Lisa
   and Bob. Aurora project plans were reviewed. There is currently
   no budget reserves for this project. Lisa will escalate to
   management. Next meeting on Tuesday.\n
-  2. Telephone Conference: ABC Corp. sales representative called
-  to discuss new printer. Promised to get us a demo by Friday.\n
-  3. Henry Miller (Handsoff Insurance): Car was totaled by tree.
-  Is looking into a loaner car. 654-2323 (tel).
-END:VJOURNAL
-END:VCALENDAR
-"""
-
-journal2 = """
-BEGIN:VCALENDAR
-VERSION:2.0
-PRODID:-//Example Corp.//CalDAV Client//EN
-BEGIN:VJOURNAL
-UID:19970901T130000Z-123405@host.com
-DTSTAMP:19970901T130000Z
-DTSTART;VALUE=DATE:19970317
-SUMMARY:Staff meeting minutes
-DESCRIPTION:1. Staff meeting: Participants include Joe\, Lisa and Bob. Aurora project plans were reviewed. There is currently no budget reserves for this project. Lisa will escalate to management. Next meeting on Tuesday.\n 2. Telephone Conference: ABC Corp. sales representative called to discuss new printer. Promised to get us a demo by Friday.\n 3. Henry Miller (Handsoff Insurance): Car was totaled by tree. Is looking into a loaner car. 654-2323 (tel).
 END:VJOURNAL
 END:VCALENDAR
 """
@@ -165,8 +163,7 @@ testcal_id = "pythoncaldav-test"
 testcal_id2 = "pythoncaldav-test2"
 
 class RepeatedFunctionalTestsBaseClass(object):
-    """
-    This is a class with functional tests (tests that goes through
+    """This is a class with functional tests (tests that goes through
     basic functionality and actively communicates with third parties)
     that we want to repeat for all configured caldav_servers.
     
@@ -180,10 +177,13 @@ class RepeatedFunctionalTestsBaseClass(object):
        user (or principal), and those are typically pre-created for each
        account.
 
-    However, iCloud is the only server where I have been denied creating a
-    calendar.  Creating a calendar through the WebUI works, and creating an
-    event through the library fails, so I don't think the problem is lack of 
+    On iCloud I've been denied creating a calendar.  Creating a
+    calendar through the WebUI works, and creating an event through
+    the library fails, so I don't think the problem is lack of
     MKCALENDAR support.
+
+    On Radicale, apparently there is some bug with MKCALENDAR, ref
+    https://github.com/Kozea/Radicale/issues/330
     """
     def setup(self):
         logging.debug("############## test setup")
@@ -219,6 +219,12 @@ class RepeatedFunctionalTestsBaseClass(object):
         Test of the propfind methods. (This is sort of redundant, since
         this is implicitly run by the setup)
         """
+        ## ResourceType MUST be defined, and SHOULD be returned on a propfind for
+        ## "allprop" if I have the permission to see it.
+        ## So, no ResourceType returned seems like a bug in bedework
+        if 'nopropfind' in self.server_params:
+            raise SkipTest("Skipping propfind test, re test suite configuration.  Perhaps the caldav server is not adhering to the standards")
+        
         ## first a raw xml propfind to the root URL
         foo = self.caldav.propfind(self.principal.url, props="""<?xml version="1.0" encoding="UTF-8"?>
 <D:propfind xmlns:D="DAV:">
@@ -240,8 +246,7 @@ class RepeatedFunctionalTestsBaseClass(object):
 
     def testProxy(self):
         if self.caldav.url.scheme == 'https':
-            logging.info("Skipping %s.testProxy as the TinyHTTPProxy implementation doesn't support https")
-            return
+            raise SkipTest("Skipping %s.testProxy as the TinyHTTPProxy implementation doesn't support https")
 
         server_address = ('127.0.0.1', 8080)
         proxy_httpd = NonThreadingHTTPServer (server_address, ProxyHandler, logging.getLogger ("TinyHTTPProxy"))
@@ -293,14 +298,15 @@ class RepeatedFunctionalTestsBaseClass(object):
         assert_equal(len(events), 0)
         events = self.principal.calendar(name="Yep", cal_id=testcal_id).events()
         ## huh ... we're quite constantly getting out a list with one item, the URL for
-	## the caldav server.  This needs to be investigated, it is surely a bug in our
-	## code.  Anyway, better to ignore it now than to have broken test code.
+        ## the caldav server.  This needs to be investigated, it is surely a bug in our
+        ## code.  Anyway, better to ignore it now than to have broken test code.
         assert_equal(len(events), 0)
         c.delete()
         
         ## verify that calendar does not exist - this breaks with zimbra :-(
+        ## (also breaks with radicale, which by default creates a new calendar)
         ## COMPATIBILITY PROBLEM - todo, look more into it
-        if 'zimbra' not in str(c.url):
+        if not 'nocalendarnotfound' in self.server_params:
             assert_raises(error.NotFoundError, self.principal.calendar(name="Yep", cal_id=testcal_id).events)
 
     def testCreateCalendarAndEvent(self):
@@ -326,17 +332,16 @@ class RepeatedFunctionalTestsBaseClass(object):
         * It will add some journal entries to it
         * It will list out all journal entries
         """
-        if '/remote.php/caldav' in str(self.caldav.url) or self.caldav.url.path.startswith('/dav/'):
+        if 'nojournal' in self.server_params:
             ## COMPATIBILITY TODO: read the RFC.  sabredav/owncloud:
             ## got the error: "This calendar only supports VEVENT,
             ## VTODO. We found a VJOURNAL".  Should probably learn
             ## that some other way.  (why doesn't make_calendar break?
             ## what does the RFC say on that?)  Same with zimbra,
             ## though different error.
-            return
+            raise SkipTest("Journal testing skipped due to test configuration")
         c = self.principal.make_calendar(name="Yep", cal_id=testcal_id, supported_calendar_component_set=['VJOURNAL'])
-        #j1 = c.add_journal(journal1) ## vobjects breaks here
-        j1 = c.add_journal(journal2)
+        j1 = c.add_journal(journal)
         journals = c.journals()
         assert_equal(len(journals), 1)
         todos = c.todos()
@@ -347,12 +352,14 @@ class RepeatedFunctionalTestsBaseClass(object):
         """
         This test demonstrates the support for task lists.
         * It will create a "task list"
-        * It will add some tasks to it
-        * It will list out all pending tasks, sorted by due date
-        * It will list out all pending tasks, sorted by priority
-        * It will complete a task
-        * It will delete a task
+        * It will add a task to it
+        * Verify the cal.todos() method
+        * Verify that cal.events() method returns nothing
         """
+        ## bedeworks does not support VTODO
+        if 'notodo' in self.server_params:
+            raise SkipTest("VTODO testing skipped due to test configuration")
+            
         ## For all servers I've tested against except Zimbra, it's
         ## possible to create a calendar and add todo-items to it.
         ## Zimbra has separate calendars and task lists, and it's not
@@ -360,44 +367,139 @@ class RepeatedFunctionalTestsBaseClass(object):
         ## tell Zimbra that the new "calendar" is a task list.  This
         ## is done though the supported_calendar_compontent_set
         ## property - hence the extra parameter here:
+        logging.info("Creating calendar Yep for tasks")
         c = self.principal.make_calendar(name="Yep", cal_id=testcal_id, supported_calendar_component_set=['VTODO'])
 
         ## add todo-item
+        logging.info("Adding todo item to calendar Yep")
         t1 = c.add_todo(todo)
 
         ## c.todos() should give a full list of todo items
+        logging.info("Fetching the full list of todo items (should be one)")
         todos = c.todos()
         assert_equal(len(todos), 1)
 
+        logging.info("Fetching the events (should be none)")
         ## c.events() should NOT return todo-items
         events = c.events()
         assert_equal(len(events), 0)
 
-	t2 = c.add_todo(todo2)
+    def testTodos(self):
+        """
+        This test will excercise the cal.todos() method, and in particular the sort_keys attribute.
+        * It will list out all pending tasks, sorted by due date
+        * It will list out all pending tasks, sorted by priority
+        """
+        ## bedeworks does not support VTODO
+        if 'notodo' in self.server_params:
+            raise SkipTest("VTODO testing skipped due to test configuration")
+            
+        c = self.principal.make_calendar(name="Yep", cal_id=testcal_id, supported_calendar_component_set=['VTODO'])
+
+        ## add todo-item
+        t1 = c.add_todo(todo)
+        t2 = c.add_todo(todo2)
         t3 = c.add_todo(todo3)
 
-	todos = c.todos()
+        todos = c.todos()
 
         assert_equal(len(todos), 3)
         uids = lambda lst: [x.instance.vtodo.uid for x in lst]
         assert_equal(uids(todos), uids([t2, t3, t1]))
 
-        todos = c.todos(sort_key='priority')
+        todos = c.todos(sort_keys=('priority',))
         pri = lambda lst: [x.instance.vtodo.priority.value for x in lst if hasattr(x.instance.vtodo, 'priority')]
         assert_equal(pri(todos), pri([t3, t2]))
 
+        todos = c.todos(sort_keys=('summary','priority',))
+        assert_equal(uids(todos), uids([t3, t2, t1]))
+
+    def testTodoDatesearch(self):
+        """
+        Let's see how the date search method works for todo events
+        """
+        ## bedeworks does not support VTODO
+        if 'notodo' in self.server_params:
+            raise SkipTest("VTODO testing skipped due to test configuration")
+            
+        c = self.principal.make_calendar(name="Yep", cal_id=testcal_id, supported_calendar_component_set=['VTODO'])
+
+        ## add todo-item
+        t1 = c.add_todo(todo)
+        t2 = c.add_todo(todo2)
+        t3 = c.add_todo(todo3)
+        t4 = c.add_todo(todo4)
+        todos = c.todos()
+        assert_equal(len(todos), 4)
+
+        notodos = c.date_search(start=datetime(1997, 4, 14), end=datetime(2015,5,14)) ## default compfilter is events
+        assert(not notodos)
+
+        ## Now, this is interesting.  2 events have dtstart set, 3 has
+        ## due set and 1 has neither due nor dtstart set.  None has
+        ## duration set.  What will a date search yield?
+        todos = c.date_search(start=datetime(1997, 4, 14), end=datetime(2015,5,14), compfilter='VTODO')
+        ## The RFCs are pretty clear on this.  rfc5545 states:
+
+        ### A "VTODO" calendar component without the "DTSTART" and "DUE" (or
+        ### "DURATION") properties specifies a to-do that will be associated
+        ### with each successive calendar date, until it is completed.
+
+        ## and RFC4791, section 9.9 also says that events without
+        ## dtstart or due should be counted.  Since we have "expand"
+        ## set, it could even imply that we should get two VTODO items
+        ## our for each day in the time range!  In any case, less than
+        ## 4 todos returned is a breach of the RFCs.
+
+        ## TODO: prod the caldav server implementators about the RFC
+        ## breakages.
+
+        ## This is probably correct, and most server implementations gives this:
+        #assert_equal(len(todos), 4)
+        ## ... but some caldav implementations yields 2 and 3:
+        assert(len(todos) >= 2)
+
+    def testTodoCompletion(self):
+        """
+        Will check that todo-items can be completed and deleted
+        """
+        ## bedeworks does not support VTODO
+        if 'notodo' in self.server_params:
+            raise SkipTest("VTODO testing skipped due to test configuration")
+            
+        c = self.principal.make_calendar(name="Yep", cal_id=testcal_id, supported_calendar_component_set=['VTODO'])
+
+        ## add todo-items
+        t1 = c.add_todo(todo)
+        t2 = c.add_todo(todo2)
+        t3 = c.add_todo(todo3)
+
+        ## There are now three todo-items at the calendar
+        todos = c.todos()
+        assert_equal(len(todos), 3)
+        
+        ## Complete one of them
         t3.complete()
-	todos = c.todos()
+        
+        ## There are now two todo-items at the calendar
+        todos = c.todos()
         assert_equal(len(todos), 2)
 
-	todos = c.todos(include_completed=True)
+        ## The historic todo-item can still be accessed
+        todos = c.todos(include_completed=True)
         assert_equal(len(todos), 3)
 
         t2.delete()
 
+        ## ... the deleted one is gone ...
         todos = c.todos(include_completed=True)
         assert_equal(len(todos), 2)
 
+        ## date search should not include completed events ... hum.  TODO, fixme.
+        #todos = c.date_search(start=datetime(1990, 4, 14), end=datetime(2015,5,14), compfilter='VTODO', hide_completed_todos=True)
+        #assert_equal(len(todos), 1)
+        
+
     def testUtf8Event(self):
         c = self.principal.make_calendar(name="Yølp", cal_id=testcal_id)
 
@@ -528,7 +630,7 @@ class RepeatedFunctionalTestsBaseClass(object):
         ## Lets try a freebusy request as well
         ## except for on my own DAViCal, it returns 500 Internal Server Error for me.  Should look more into that.  TODO.
         if 'calendar.bekkenstenveien53c.oslo' in str(self.caldav.url):
-            return
+            raise SkipTest("TEMP/TODO/KLUDGE - skipping this test as for now")
         freebusy = c.freebusy_request(datetime(2007,7,13,17,00,00),
                                       datetime(2007,7,15,17,00,00))
         ## TODO: assert something more complex on the return object
@@ -541,6 +643,8 @@ class RepeatedFunctionalTestsBaseClass(object):
         library per se.  How will it behave if we serve it a recurring
         event?
         """
+        if 'norecurring' in self.server_params:
+            raise SkipTest("recurring date search test skipped due to test configuration")
         c = self.principal.make_calendar(name="Yep", cal_id=testcal_id)
 
         ## evr is a yearly event starting at 1997-02-11
